
/****************************************************************************
 National Instruments Corporation.

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 ****************************************************************************/

/*
 * Si514 Clock kernel driver
 */



/* si514 data sheet
https://www.silabs.co/Support%20Documents/TechnicalDocs/Si514.pdf
 */

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/module.h>
#include <linux/i2c.h> /* i2c device communications */
#include <linux/kobject.h>
#include <linux/sysfs.h>

#include <linux/slab.h>
#include <linux/device.h>
#include <asm/page.h>

#include <linux/types.h>
#include <generated/uapi/linux/version.h>
/* enabling floating point math */
#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 2, 0)
	#include <asm/i387.h>
#else
	#include <asm/fpu/api.h>
#endif
#include <linux/moduleparam.h>

/****************************************************************************
 * Si514 Regmap *
 ****************************************************************************/

/* Si514 Registers Offset */
#define SI514_REG_OFFSET_M_FRAC1  5
#define SI514_REG_OFFSET_M_FFRAC2  6
#define SI514_REG_OFFSET_M_FRAC3  7
#define SI514_REG_OFFSET_M_INT_FRACT 8
#define SI514_REG_OFFSET_M_INT  9
#define SI514_REG_OFFSET_HS_DIV  10
#define SI514_REG_OFFSET_LS_HS_DIV 11

/* Register Masks */
#define SI514_RESET_RST  BIT(7)
#define SI514_CONTROL_FCAL BIT(0)
#define SI514_CONTROL_OE BIT(2)

/* Default adjustment range comes from the 30000pbb max
** stability of the 514 part itself (it needs to be at least
** double that so that two 514-based devices at opposite
** ranges will be able to sync to each other) plus a 2ppm
** extra tuning range for the servo. */
#define SI514_DEFAULT_ADJ_PPB 62000


/* Fixed-frequency Crystal, 29-bit resolution */
static const unsigned int si514_FXO = 31980000;
/* 536870912 dec,  Scaling factor for the 29-bit */
static const unsigned int si514_29bit_val = 0x20000000;
static const unsigned int si514_Ppb_divisor = 1000000000;
static const int fraction_scale = 1000;

/* Multiplier/divider settings */
struct tSi514_muldiv {
	u32 m_frac; /* 29-bit Fractional part of multiplier M */
	u8 m_int; /* Integer part of multiplier M, 65..78 */
	u8 ls_div_bits; /* 2nd divider, as 2^x */
	u16 hs_div; /* 1st divider, must be even and 10<=x<=1022 */
};

/*
 * Default values for 100Mhz operation (from powerup state of 100Mhz
 * factory part)
 */
static const struct tSi514_muldiv k100MhzReferenceHWSettings = {
	.m_frac = 425736282,
	.m_int = 68,
	.ls_div_bits = 0,
	.hs_div = 22
};

/* Default values for 25Mhz operation (from Si514 datasheet table) */
static const struct tSi514_muldiv k25MhzReferenceHWSettings = {
	.m_frac = 357578187,
	.m_int = 65,
	.ls_div_bits = 0,
	.hs_div = 84
};

/** Device State (client specific and additional data) **/
struct si514_data {
	struct i2c_client *i2c_client;
	struct mutex lock;
	struct tSi514_muldiv current_hw_settings;
	const struct tSi514_muldiv *reference_hw_settings;
};

/****************************************************************************
 * Functions Declaration *
 ****************************************************************************/
static int si514_probe(struct i2c_client *client,
	const struct i2c_device_id *id);
static int si514_remove(struct i2c_client *client);

/* ------------------- Device Specific functions --------------------- */
static int si514_initial_configuration(struct i2c_client *client);
static int si514_shift_ppb(struct i2c_client *client,
	s32 desired_shift_in_ppb_int,
	s32 desired_shift_in_ppb_fract);
static int si514_update_multidiv_soft_copy(struct i2c_client *client);
static int si514_set_small_tune(struct i2c_client *client,
	const struct tSi514_muldiv *new_hw_settings);
static void compute_current_ppb_shift(const struct i2c_client *client,
	s32 *ppb_int_part, s32 *ppb_fract_part);
static int si514_read_status(const struct i2c_client *client,
	struct tSi514_muldiv *settings);
static int si514_set_status(const struct i2c_client *client,
	const struct tSi514_muldiv *settings);
static int si514_trigger_status(const struct i2c_client *client,
	const struct tSi514_muldiv *settings);
static int si514_match_hw_status(const struct tSi514_muldiv *settings01,
	const struct tSi514_muldiv *settings02);
static int si514_match_hw_trigger_status(const struct tSi514_muldiv *settings01,
	const struct tSi514_muldiv *settings02);

/**********************
 * Device ID table *
 **********************/
static const struct i2c_device_id si514_idtable[] = {
	{"si514", 0},
	{}
};

MODULE_DEVICE_TABLE(i2c, si514_idtable);

/***********************
 * Device Driver *
 ***********************/
static struct i2c_driver si514_driver = {
	.driver = {
		.name = "si514",
		.owner = THIS_MODULE,
	},
	.id_table = si514_idtable,
	.probe = si514_probe,
	.remove = si514_remove,
};


/*** Helper Macros (use when fpu is enabled) ***/
#define SI514_GET_CURR_MUTLIPLIER(settings) (double)(((settings)->m_int) + \
	(((double)((settings)->m_frac) / (si514_29bit_val))))

#define SI514_GET_CURR_FREQ_HZ(settings)  \
	(double)(((SI514_GET_CURR_MUTLIPLIER(settings)) * \
	(double) (si514_FXO)) / (((settings)->hs_div) * \
	(BIT((settings)->ls_div_bits))));

/************************* Device Attributes *****************************/
static ssize_t si514_show_ppb_offset(struct device *dev,
	struct device_attribute *dev_attr,
	char *buf)
{
	s32 ppb_int = 0;
	s32 ppb_fract = 0;
	struct i2c_client *client;

	client = to_i2c_client(dev);

	compute_current_ppb_shift(client, &ppb_int, &ppb_fract);

	/* The return value is float, represented as two long numbers
	 * seperated (formatted) by dot.
	 * the first one is The integer part and the second one is
	 * the fraction as a part of 1000.
	*/

	return scnprintf(buf, PAGE_SIZE, "%ld.%ld\n", (long) ppb_int,
			(long) ppb_fract);
}

static ssize_t si514_store_ppb_offset(struct device *dev,
	struct device_attribute *dev_attr,
	const char *buf, size_t count)
{
	s32 shift_ppb_int;
	s32 shift_ppb_fract;
	int ret;
	struct i2c_client *client;
	client = to_i2c_client(dev);

	/* The expected received value is float, represented as two long
	 * numbers seperated (formatted) by dot.
	 * the first one is The integer part and the second one is
	 * the fraction as a part of 1000.
	*/

	ret = sscanf(buf, "%ld.%ld", (long *) &shift_ppb_int,
			(long *) &shift_ppb_fract);

	if (ret != 2) {
		dev_warn(dev, "\n %s : Failed to read input parameter",
			 __func__);
		return -EIO;
	}

	ret = si514_shift_ppb(client, shift_ppb_int, shift_ppb_fract);
	if (ret < 0)
		return ret;
		/* User mode code wil detect that this call failed */

	return count;
}

static DEVICE_ATTR(ppb_offset, S_IWUSR | S_IWGRP | S_IRUGO,
	si514_show_ppb_offset, si514_store_ppb_offset);

/*----------------------------------------------------------------------*/
static ssize_t si514_show_fraction_scale(struct device *dev,
	struct device_attribute *dev_attr,
	char *buf)
{
return sprintf(buf, "%d\n", fraction_scale);
}
static DEVICE_ATTR(fraction_scale, S_IRUGO, si514_show_fraction_scale, NULL);
/*-----------------------------------------------------------------------*/

static struct attribute *si514_attributes[] = {
	&dev_attr_ppb_offset.attr, /*RW*/
	&dev_attr_fraction_scale.attr, /*R*/
	NULL
};

static struct attribute_group si514_attr_grp = {
	.attrs = si514_attributes,
};
/*///////////////////////////////////////////////////////////////////////////*/
static s32 max_allowed_ppb = SI514_DEFAULT_ADJ_PPB;
module_param(max_allowed_ppb, int, 0644);
MODULE_PARM_DESC(max_allowed_ppb, "Maximum allowed ppb tuning value");

module_i2c_driver(si514_driver);

/*///////////////////////////////////////////////////////////////////////////*/

/*** Probe ***/
static int si514_probe(struct i2c_client *client,
	const struct i2c_device_id *id)
{
	int result;
	struct si514_data *si514_state_data;

	/* Initialize the driver data object */
	si514_state_data = NULL;
	si514_state_data = devm_kzalloc(&client->dev,
		sizeof(*si514_state_data), GFP_KERNEL);
	if (si514_state_data == NULL) {
		dev_err(&client->dev,
		"%s : failed to create driver data object\n", __func__);
		return -ENOMEM;
	}

	si514_state_data->i2c_client = client;
	i2c_set_clientdata(client, si514_state_data);

	/* Initialize Mutex */
	mutex_init(&si514_state_data->lock);


	/* Register device sttributes to the sysfs */
	result = sysfs_create_group(&client->dev.kobj, &si514_attr_grp);
	if (result)
		goto error_mem;

	/* Device Initial Configuration */
	result = si514_initial_configuration(client);
	if (result < 0) {
		/* Usually this failure due to HW communication */
		dev_err(&client->dev, "\n Failed initial configurations! ");
		goto error_class;
	}

	return 0;

error_class:
	sysfs_remove_group(&client->dev.kobj, &si514_attr_grp);

error_mem:
	devm_kfree(&client->dev, si514_state_data);

	return result;
}

/*///////////////////////////////////////////////////////////////////////////*/
static int si514_remove(struct i2c_client *client)
{
	struct si514_data *dev_data;

	sysfs_remove_group(&client->dev.kobj, &si514_attr_grp);
	dev_data = i2c_get_clientdata(client);
	devm_kfree(&client->dev, dev_data);
	return 0;
}

/*///////////////////////////////////////////////////////////////////////////*/

static int si514_initial_configuration(struct i2c_client *client)
{
	int result;
	struct si514_data *dev_data_ptr;

	dev_data_ptr = i2c_get_clientdata(client);

	result = si514_update_multidiv_soft_copy(client);
	if (result < 0) {
		dev_err(&client->dev, "\n %s :failed to read HW registers\n",
			__func__);
		return result;
	}

	/*
	 * Detect which reference values we should be using based on
	 * hs_div setting. Since we do not adjust it in this driver,
	 * it should remain what it powered up with. The two reference
	 * values (25 and 100Mhz) this driver expects to deal with luckily
	 * have different hs_div values
	 */
	if (dev_data_ptr->current_hw_settings.hs_div ==
		k100MhzReferenceHWSettings.hs_div) {
		dev_data_ptr->reference_hw_settings =
			&k100MhzReferenceHWSettings;
	} else if (dev_data_ptr->current_hw_settings.hs_div ==
		k25MhzReferenceHWSettings.hs_div) {
		dev_data_ptr->reference_hw_settings =
			&k25MhzReferenceHWSettings;
	} else {
		dev_err(&client->dev,
		"%s :unknown value for hs_div\n", __func__);
		return -EIO;
	}

	/* Shift the frequency to default */
	result = si514_set_small_tune(client,
		dev_data_ptr->reference_hw_settings);
	if (result < 0)
		return result; /* fail initialization */

	/*
	 * Read the HW to validate we are in good state before moving
	 * forward
	 */
	result = si514_update_multidiv_soft_copy(client);
	if (result < 0) {
		dev_err(&client->dev, "\n %s :failed to read HW registers\n",
			__func__);
		return result;
	}

	/* at this point we are expecting the state of the HW registers
	is the same as the reference HW settings. Validate that */
	if (dev_data_ptr->reference_hw_settings->m_frac !=
		dev_data_ptr->current_hw_settings.m_frac ||
		dev_data_ptr->reference_hw_settings->m_int !=
		dev_data_ptr->current_hw_settings.m_int) {
		dev_warn(&client->dev,
			"\n %s : Reference settings doesn't match the initial HW state\n",
			__func__);
		return -EIO;
	}

return 0;
}

/*///////////////////////////////////////////////////////////////////////////*/

static int si514_shift_ppb(struct i2c_client *client,
	s32 desired_shift_in_ppb_int,
	s32 desired_shift_in_ppb_fract)
{
	int result;
	u16 new_int;
	u32 new_frac;
	struct tSi514_muldiv new_freq_settings;
	struct si514_data *dev_data;
	double desired_shift_ppb;
	double desired_frequency_shift_hz_from_ref_freq;
	double reference_frequency_hz;
	double desired_frequency_hz;
	double reference_m;
	double div_freq_d;
	double new_m;
	double si514_m_max;
	double si514_m_min;
	double new_freq_hz;

	dev_data = i2c_get_clientdata(client);

	kernel_fpu_begin();
	/* Intel x86_64 platform specific.
	Used because we have 29 fraction that need total of 128-bit to
	effectively divide and multiply.
	There is away to do that but it need time to get it to work,
	therefore, we are enabling the floating point math in the kernel at the
	moment and this driver will only be running on this platform.
	We should follow up on this to enable the fix point math.*/

	desired_shift_ppb = desired_shift_in_ppb_int +
				((double) desired_shift_in_ppb_fract /
				fraction_scale);

	if (desired_shift_ppb > max_allowed_ppb ||
		desired_shift_ppb < -max_allowed_ppb) {
		dev_warn(&client->dev,
			"\n Desired PPB INT = %ld, is out of range",
			(long) desired_shift_ppb);
		dev_warn(&client->dev,
			"\n Max PPB INT = %d, is out of range",
			max_allowed_ppb);
		kernel_fpu_end();
		return -ERANGE;
	}

	/* Compute the frequency shift from the initial state frequency */
	reference_frequency_hz =
		SI514_GET_CURR_FREQ_HZ(dev_data->reference_hw_settings);

	desired_frequency_shift_hz_from_ref_freq = (double) ((double)
		((double) reference_frequency_hz * desired_shift_ppb) /
		si514_Ppb_divisor);
	desired_frequency_hz = reference_frequency_hz +
		desired_frequency_shift_hz_from_ref_freq;

	/*------ Shift Frequency -------*/
	reference_m =
		SI514_GET_CURR_MUTLIPLIER(dev_data->reference_hw_settings);

	div_freq_d = (double) desired_frequency_hz / reference_frequency_hz;
	new_m = reference_m * div_freq_d;

	/* Taking the integer part only */
	new_int = (u16) new_m; /*(u32) (Mnew >> 29); */
	/* taking the fraction part only (29-bit resolution) */
	new_frac = (new_m - new_int)*si514_29bit_val;

	/* Range Validation, check before writing to the HW */
	si514_m_max = 78.17385866; /* From data sheet page #12 */
	si514_m_min = 65.04065041; /* From data sheet page #12 */
	if (new_m > si514_m_max || new_m < si514_m_min) {
		dev_warn(&client->dev,
			"\n Failed : Out of range, computed M = %u.%u\n",
			new_int, new_frac);
		kernel_fpu_end();
		return -EIO;
	}

	new_freq_settings = *dev_data->reference_hw_settings;
	new_freq_settings.m_frac = new_frac;
	new_freq_settings.m_int = new_int;

	/* Debugging */
	new_freq_hz = SI514_GET_CURR_FREQ_HZ(&new_freq_settings);
	/*-------------*/

	kernel_fpu_end();

	/* Shift the frequency */
	result = si514_set_small_tune(client, &new_freq_settings);

	return result;
}

/*///////////////////////////////////////////////////////////////////////////*/
static int si514_update_multidiv_soft_copy(struct i2c_client *client)
{
	struct si514_data *dev_data;
	int result;

	result = 0;
	dev_data = i2c_get_clientdata(client);

	result = si514_read_status(client, &dev_data->current_hw_settings);

	return result;
}

/*///////////////////////////////////////////////////////////////////////////*/

/* Configure the clock settings on the hardware */
static int si514_set_small_tune(struct i2c_client *client,
	const struct tSi514_muldiv *new_hw_settings)
{
	/* The si514 controls many clocks on the system. We
	   need to make sure that we don't put it in bad state as much
	   as we can.
	   The logic in this function as the following:
	   - write the desired values to the HW, but with out triggering the
		frequency update.
	   - read back and make sure they match what we already wrote.
	   - If it match, then trigger the frequency change. If not,
		try to write one more time (just to avoid any short and
		temporary noise on the bus "if any").
	   - After trigging the change, read the HW status one more time
		to make sure it match all the values have been written.
	   - if it didn't match, then abort and dont update the current
		HW settings, the user code can detect that something went
		wrong as the requested frequency didn't honored.
	*/

	int result;
	struct tSi514_muldiv settings;
	struct si514_data *dev_data;

	/* update the HW and trigger the changes */
	result = si514_set_status(client, new_hw_settings);
	if (result < 0)
		return result;

	/* read back status and make sure the status as has been set */
	result = si514_read_status(client, &settings);
	if (result < 0)
		return result;

	if (si514_match_hw_status(new_hw_settings, &settings) < 0) {
		/* data write fail. Try one more time */
		result = si514_set_status(client, new_hw_settings);
		if (result < 0)
			return result;

		if (si514_match_hw_status(new_hw_settings, &settings) < 0) {
			dev_err(&client->dev, "\n Failed to tune new frequency. IO error ");
			return -EIO;
		}
	}

	/* Trigger the frequency update */
	result = si514_trigger_status(client, new_hw_settings);
	if (result < 0)
		return result;

	/* read back status and make sure the status as has been set */
	result = si514_read_status(client, &settings);
	if (result < 0)
		return result;

	if (si514_match_hw_trigger_status(new_hw_settings, &settings) < 0) {
		dev_err(&client->dev, "\n Failed to tune new frequency. IO error ");
		return -EIO;
	}

	/* Reaching this point, means that the frequency update completed
	 * successfuly. Updat ethe current HW settings */
	 dev_data = i2c_get_clientdata(client);
	 dev_data->current_hw_settings = *new_hw_settings;

	return result;
}

/*///////////////////////////////////////////////////////////////////////////*/

void compute_current_ppb_shift(const struct i2c_client *client,
	s32 *ppb_int_part, s32 *ppb_fract_part)
{
	struct si514_data *dev_data;
	double ref_freq_hz;
	double freq_shift;
	double current_freq_hz;
	double ppb_shifted;

	dev_data = i2c_get_clientdata(client);

	kernel_fpu_begin();

	/* Compute Reference frequency from reference HW state */
	ref_freq_hz = SI514_GET_CURR_FREQ_HZ(dev_data->reference_hw_settings);

	mutex_lock(&dev_data->lock);
	/* Compute Current frequency from current HW state */
	current_freq_hz =
		SI514_GET_CURR_FREQ_HZ(&dev_data->current_hw_settings);

	/* Compute the frequency shift in ppb units */
	freq_shift = current_freq_hz - ref_freq_hz;

	ppb_shifted = (double) ((double) freq_shift / ref_freq_hz)
			*si514_Ppb_divisor;

	/* Update the shift in ppb */
	*ppb_int_part = (s32) ppb_shifted;
	*ppb_fract_part = (ppb_shifted - *ppb_int_part)*fraction_scale;

	kernel_fpu_end();

	mutex_unlock(&dev_data->lock);
}

/*///////////////////////////////////////////////////////////////////////////*/

static int si514_set_status(const struct i2c_client *client,
	const struct tSi514_muldiv *settings)
{
	struct si514_data *dev_data;
	int result;
	/* Write all fine-tuning registers except the trigger register.
	the trigger registyer is at SI514_REG_OFFSET_M_INT offset,
	which cause the device to updat the frequency */
	u8 reg_buff[4];

	result = 0;
	dev_data = i2c_get_clientdata(client);

	reg_buff[0] = settings->m_frac & 0xff;
	reg_buff[1] = (settings->m_frac >> 8) & 0xff;
	reg_buff[2] = (settings->m_frac >> 16) & 0xff;
	reg_buff[3] = ((settings->m_frac >> 24) |
			(settings->m_int << 5)) & 0xff;

	mutex_lock(&dev_data->lock);

	result = i2c_smbus_write_i2c_block_data(client,
		(u8) SI514_REG_OFFSET_M_FRAC1, (u8) ARRAY_SIZE(reg_buff),
		reg_buff);

	if (result < 0) {
		/* Don't update the current HW settings
		 * At this stage the user mod code will detect that there
		 * is something wrong and will take action depends on that.
		 * At the driver side, this is a HW issue that we cant do
		 * too much about.*/
		dev_err(&client->dev, "%s : failed to write registers\n",
			__func__);
	}

	mutex_unlock(&dev_data->lock);

	return result;
}

/*///////////////////////////////////////////////////////////////////////////*/

static int si514_trigger_status(const struct i2c_client *client,
	const struct tSi514_muldiv *settings)
{
	struct si514_data *dev_data;
	u8 m_int;
	int result;

	dev_data = i2c_get_clientdata(client);

	m_int = (settings->m_int >> 3) & 0xff;

	mutex_lock(&dev_data->lock);

	result = i2c_smbus_write_byte_data(client, (u8) SI514_REG_OFFSET_M_INT,
		(u8) m_int);

	if (result < 0) {
		dev_err(&client->dev, "%s : failed to write trigger registers\n",
			__func__);
		goto err;
	}

	udelay(100); /* Small tuning settling time. 100 micro second */

err:
	mutex_unlock(&dev_data->lock);
	return result;
}

/*///////////////////////////////////////////////////////////////////////////*/

static int si514_read_status(const struct i2c_client *client,
	struct tSi514_muldiv *settings)
{
	int result;
	struct si514_data *dev_data;
	u8 reg_buff[7];

	dev_data = i2c_get_clientdata(client);

	mutex_lock(&dev_data->lock);

	result = i2c_smbus_read_i2c_block_data(client,
		(u8) SI514_REG_OFFSET_M_FRAC1, (u8) ARRAY_SIZE(reg_buff),
		reg_buff);

	if (result < 0) {
		dev_err(&client->dev, "%s : failed to read registers\n",
			__func__);
		goto release;
	}

	settings->m_frac = reg_buff[0] | ((u32) reg_buff[1] << 8) |
		((u32) reg_buff[2] << 16) |
		((u32) (reg_buff[3] & 0x1F) << 24);
	settings->m_int = ((u16) (reg_buff[4] & 0x3f) << 3) |
		(reg_buff[3] >> 5);
	settings->ls_div_bits = (reg_buff[6] >> 4) & 0x07;
	settings->hs_div = ((u16) (reg_buff[6] & 0x03) << 8) | reg_buff[5];

release:
	mutex_unlock(&dev_data->lock);
	return result;
}

/*///////////////////////////////////////////////////////////////////////////*/

int si514_match_hw_status(const struct tSi514_muldiv *settings01,
	const struct tSi514_muldiv *settings02)
{
	if (settings01->m_frac == settings02->m_frac)
		return 0;

	return -1;
}

/*///////////////////////////////////////////////////////////////////////////*/

int si514_match_hw_trigger_status(const struct tSi514_muldiv *settings01,
	const struct tSi514_muldiv *settings02)
{
	if (settings01->m_int == settings02->m_int)
		return 0;

	return -1;
}

/*///////////////////////////////////////////////////////////////////////////*/

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Driver for Si514 Oscillators, I2C Driver");
MODULE_AUTHOR("Bassam Alsanie <bassam.alsanie@ni.com>");
